/*
 * ======================================================================
 * key_sniff.c
 * Simple example how to use keyboard sniffing interface in the kernel.
 * Copyright 1998 Roger Butenuth
 * butenuth@uni-paderborn.de
 * ======================================================================
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <linux/kd.h>
#include <linux/keyboard.h>
#include <linux/kbd_sniff.h>


int cons_fd;			/* for key to ascii translation */
int vc_fd;


/*
 *----------------------------------------------------------------------
 * Convert given shift state and keycode to ascii. 
 * Return 0 whenever there is no meaningful translation.
 *----------------------------------------------------------------------
 */
int key_to_ascii(int shift_state, int keycode)
{
    struct kbentry key;
    int            type, sym;
    int            ascii = 0;

    if (keycode & 0x80)
	return 0;		/* ignore key release */

    key.kb_table = shift_state & 0xff;
    key.kb_index = keycode;
    key.kb_value = 0;

    if (ioctl(cons_fd, KDGKBENT, &key) < 0) {
	perror("get key entry ioctl");
	return 0;
    }
    if (key.kb_value != K_HOLE) {
	type = KTYP(key.kb_value) & 0x0f;
	sym  = KVAL(key.kb_value);
	switch (type) {
	case KT_LETTER:
	    if ((shift_state & SNIFF_CAPSLOCK) != 0)
		ascii = sym - 'a' + 'A';
	    else
		ascii = sym;
	    break;
	case KT_LATIN:
	    ascii = sym;
	    break;
	case KT_ASCII:
	case KT_FN:
	case KT_SPEC:
	case KT_PAD:
	case KT_DEAD:
	case KT_CONS:
	case KT_CUR:
	case KT_SHIFT:
	case KT_META:
	case KT_LOCK:
	case KT_SLOCK:
	    break;
	}
    }
    return ascii;
}


/*
 *----------------------------------------------------------------------
 * Initialize keyboard sniffing. Magic key combination is
 * "left shift + right shift" to enter filter mode, "space" to leave
 * filter mode. In filter mode all keypresses are not given to other
 * application, only the sniffer gets this keypresses unless filter
 * mode is left.
 *----------------------------------------------------------------------
 */
int vc_init(void)
{
    struct sniff_ioctl_set set;

    vc_fd = open("/dev/vcsa0", O_RDWR,0);
    if (vc_fd < 0) {
	perror("open vc");
	return 1;
    }

    set.hot_key_enable = 1;
    set.nr_hot_keys = 2;
    set.hot_key[0]  = 42;		/* left shift */
    set.hot_key[1]  = 54;		/* right shift */
    set.disable_key = 57;		/* space */

    if (ioctl(vc_fd, KBD_SNIFF_SET, &set) < 0)
	perror("ioctl set");

    return 0;
}


/*
 *----------------------------------------------------------------------
 * Disable sniffing mode.
 *----------------------------------------------------------------------
 */
int vc_deinit(void)
{
    struct sniff_ioctl_set set;

    set.hot_key_enable = 0;
    if (ioctl(vc_fd, KBD_SNIFF_SET, &set) < 0)
	perror("ioctl set");

    close(vc_fd);

    return 0;
}


/*
 *----------------------------------------------------------------------
 * Set keyboard to sniff mode, print 10 keypresses, reset keyboard,
 * terminate.
 *----------------------------------------------------------------------
 */
int main(int argc, char *argv[])
{
    struct sniff_ioctl_get get;
    int                    i, j;

    cons_fd = open("/dev/console", O_RDWR,0);
    if (cons_fd < 0) {
	perror("open console");
	return 1;
    }
    vc_init();

    for (i = 0; i < 10; i++) {
	/*
	 * Get the keypresses blocking (polling is possible, too),
	 * reset counter.
	 */
	get.blocking = 1;
	get.count    = 0;
	assert(ioctl(vc_fd, KBD_SNIFF_GET, &get) >= 0);
	for (j = 0; j < get.count; j++) {
	    int shift_state = get.event[j].shift_state;
	    int keycode     = get.event[j].keycode;
	    printf("shift_state = 0x%x\n", shift_state);
	    printf("keycode     = 0x%x\n", keycode);
	    printf("ascii       = <%c>\n", key_to_ascii(shift_state, keycode));
	}
    }
    

    close(cons_fd);
    vc_deinit();
    return 0;
}




